// ChildView.cpp : implementation of the CChildView class
//

#include "stdafx.h"
#include "dabble.h"
#include "ChildView.h"
#include "NewGame.h"
#include "DlgSettings.h"
#include <math.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#pragma warning(disable:4244)

/////////////////////////////////////////////////////////////////////////////
// CChildView

CChildView::CChildView()
{
	m_game = NULL;
	m_defaultHeight = 6;
	m_defaultWidth = 6;
	m_defaultDepth = 20;
	m_defaultTimeLimit = 5;
	m_defaultGameLimit = -5;
	m_autoPlay = 1;
	m_computer = -1;
	m_internals = 0;
	m_dir = dir_invalid;
	m_showBeta = 0;

	m_name1 = "Dodie";
	m_name2 = "Evie";
}

CChildView::~CChildView()
{
	if (m_game)
		delete m_game;
}


BEGIN_MESSAGE_MAP(CChildView,CWnd )
	//{{AFX_MSG_MAP(CChildView)
	ON_WM_PAINT()
	ON_COMMAND(ID_NEW_GAME, OnNewGame)
	ON_WM_MOUSEMOVE()
	ON_WM_LBUTTONDOWN()
	ON_COMMAND(ID_MYMOVE, OnMymove)
	ON_COMMAND(ID_SETTINGS, OnSettings)
	ON_COMMAND(ID_EVAL, OnEval)
	ON_COMMAND(ID_UNDO, OnUndo)
	ON_WM_CHAR()
	ON_COMMAND(ID_AUTO_PLAY, OnAutoPlay)
	ON_UPDATE_COMMAND_UI(ID_AUTO_PLAY, OnUpdateAutoPlay)
	ON_COMMAND(ID_INTERNALS, OnInternals)
	ON_UPDATE_COMMAND_UI(ID_INTERNALS, OnUpdateInternals)
	ON_UPDATE_COMMAND_UI(ID_EVAL, OnUpdateEval)
	ON_COMMAND(ID_FORWARD, OnForward)
	ON_UPDATE_COMMAND_UI(ID_FORWARD, OnUpdateForward)
	ON_UPDATE_COMMAND_UI(ID_UNDO, OnUpdateUndo)
	ON_UPDATE_COMMAND_UI(ID_MYMOVE, OnUpdateMymove)
	ON_COMMAND(ID_FILE_OPEN, OnFileOpen)
	ON_COMMAND(ID_FILE_SAVE, OnFileSave)
	ON_UPDATE_COMMAND_UI(ID_STOP, OnUpdateStop)
	ON_COMMAND(ID_STOP, OnStop)
	ON_UPDATE_COMMAND_UI(ID_NEW_GAME, OnUpdateNewGame)
	ON_COMMAND(ID_UNPAUSE, OnUnpause)
	ON_COMMAND(ID_SHOW_BETA, OnShowBeta)
	ON_UPDATE_COMMAND_UI(ID_SHOW_BETA, OnUpdateShowBeta)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


/////////////////////////////////////////////////////////////////////////////
// CChildView message handlers

BOOL CChildView::PreCreateWindow(CREATESTRUCT& cs) 
{
	if (!CWnd::PreCreateWindow(cs))
		return FALSE;

	cs.dwExStyle |= WS_EX_CLIENTEDGE;
	cs.style &= ~WS_BORDER;
	cs.lpszClass = AfxRegisterWndClass(CS_HREDRAW|CS_VREDRAW|CS_DBLCLKS, 
		::LoadCursor(NULL, IDC_ARROW), HBRUSH(COLOR_WINDOW+1), NULL);

	return TRUE;
}

void CChildView::GetCoords (int &x0, int &y0, int &side)
{
	int temp;

	CRect rect;
	GetClientRect(&rect);

	side = rect.right * 0.95 / (m_game->width + 1);
	temp = rect.bottom * 0.95 / (m_game->height + 1);
	if (temp < side)
		side = temp;

	x0 = (rect.right - (m_game->width + 1) * side) / 2;
	y0 = (rect.bottom - (m_game->height + 1) * side) / 2;
}

void CChildView::OnPaint() 
{
	CPaintDC dc(this); // device context for painting
	
	CRect rect;
	GetClientRect(&rect);
	dc.FillSolidRect(&rect, GetSysColor(COLOR_WINDOW));

	if (!m_game)
		return;

	int side, x0, y0;
	int score[2];

	GetCoords(x0, y0, side);

	int i, j, x, y;

	CPen *oldpen = dc.GetCurrentPen();
	CBrush *oldbrush = dc.GetCurrentBrush();

	dc.SelectStockObject(BLACK_BRUSH);
	dc.SelectStockObject(BLACK_PEN);

	for (i = 0 ; i < m_game->width ; i++)
	{
		for (j = 0 ; j < m_game->height ; j++)
		{
			x = x0 + (i+1) * side;
			y = y0 + (j+1) * side;
			dc.Ellipse(x - 2, y - 2, x + 2, y + 2);

			if (i < m_game->width - 1)
				DrawLine(i+1, j, dir_horizontal);
			if (j < m_game->height - 1)
				DrawLine(i, j+1, dir_vertical);
		}
	}

	score[0] = score[1] = 0;
	for (i = 1 ; i < m_game->width ; i++)
	{
		for (j = 1 ; j < m_game->height ; j++)
		{
			if (!m_game->square[i][j].degree)
			{
				x = x0 + i * side;
				y = y0 + j * side;
				CRect rect(x, y, x+side, y+side);
				if (m_game->square[i][j].owner)
					dc.DrawText(m_name2, 1, &rect, DT_CENTER | DT_VCENTER | DT_SINGLELINE);
				else
					dc.DrawText(m_name1, 1 , &rect, DT_CENTER | DT_VCENTER | DT_SINGLELINE);
				score[m_game->square[i][j].owner]++;
			}
		}
	}

	dc.SelectObject(oldpen);
	dc.SelectObject(oldbrush);

	if (m_internals)
	{
		CPen *pen = new CPen(PS_SOLID, 0, RGB(255, 0, 0));
		CBrush *brush = new CBrush(RGB(255, 0, 0));
		dc.SelectObject(pen);
		dc.SelectObject(brush);

		for (i = 0 ; i < m_game->numfrags ; i++)
		{
			Node *node;
			Edge *edge;

			node = m_game->frag[i];
			edge = NULL;
			while (node)
			{
				x = x0 + node->x * side + side / 2;
				y = y0 + node->y * side + side / 2;
				dc.Ellipse(x - 3, y - 3, x + 3, y + 3);
				edge = node->NextEdge(edge);
				if (edge)
				{
					int inc = node->NextFrag(edge) ? side : side * 3 / 4;
					dc.MoveTo(x, y);
					if (edge == node->edge[0])
						y -= inc;
					else if (edge == node->edge[1])
						y += inc;
					else if (edge == node->edge[2])
						x -= inc;
					else
						x += inc;
					dc.LineTo(x, y);
				}
				node = node->NextFrag(edge);
			}
		}

		dc.SelectObject(oldpen);
		dc.SelectObject(oldbrush);
		delete pen;
		delete brush;

		/*
		char buff[2];
		buff[1] = 0;
		for (i = 0 ; i <= m_game->width ; i++)
		{
			x = x0 + i * side;
			CRect rect(x, y0, x+side, y0+side);
			buff[0] = i + '0';
			dc.DrawText(buff, &rect, DT_CENTER | DT_VCENTER | DT_SINGLELINE);
		}
		for (i = 0 ; i <= m_game->height ; i++)
		{
			y = y0 + i * side;
			CRect rect(x0, y, x0+side, y+side);
			buff[0] = i + '0';
			dc.DrawText(buff, &rect, DT_CENTER | DT_VCENTER | DT_SINGLELINE);
		}
		*/
	}

	char buff[128];

	// score
	dc.SetTextColor(RGB(0, 128, 0));
	dc.TextOut(32, 8, "Score");
	if (m_game->nummoves & 1)
		dc.SetTextColor(RGB(0, 0, 0));
	else
		dc.SetTextColor(RGB(128, 0, 0));
	sprintf(buff, "%s%s:", m_computer == 0 ? "*" : "", m_name1);
	dc.TextOut(8, 28, buff);
	sprintf(buff, "%d", score[0]);
	dc.TextOut(88, 28, buff);
	if (m_game->nummoves & 1)
		dc.SetTextColor(RGB(128, 0, 0));
	else
		dc.SetTextColor(RGB(0, 0, 0));
	sprintf(buff, "%s%s:", m_computer == 1 ? "*" : "", m_name2);
	dc.TextOut(8, 44, buff);
	sprintf(buff, "%d", score[1]);
	dc.TextOut(88, 44, buff);

	// search
	dc.SetTextColor(RGB(0, 120, 0));
	dc.TextOut(28, 76, "Search");
	dc.SetTextColor(RGB(0, 0, 0));
	sprintf(buff, "%d", m_game->last_depth);
	dc.TextOut(8, 96, "Depth:");
	dc.TextOut(88, 96, buff);
	sprintf(buff, "%d", m_game->last_leaves);
	dc.TextOut(8, 112, "Leaves:");
	dc.TextOut(88, 112, buff);

	y = 0;
	if (m_showBeta)
	{
		y = 16;
		sprintf(buff, "%d", m_game->last_beta);
		dc.TextOut(8, 128, "Beta:");
		dc.TextOut(88, 128, buff);
	}

	if (m_game->thinking)
	{
		dc.SetTextColor(RGB(0, 0, 128));
		dc.TextOut(20, 132+y, "Thinking...");
		dc.SetTextColor(RGB(0, 0, 0));
	}

	// number of moves
	sprintf(buff, "%d", m_game->nummoves);
	dc.TextOut(8, 164+y, "Moves:");
	dc.TextOut(88, 164+y, buff);

	// debug output
	if (m_game->m_pause)
		dc.TextOut(200, 8, m_game->msg);

	m_dir = dir_invalid;

	// Do not call CWnd::OnPaint() for painting messages
}


void CChildView::OnNewGame() 
{
	NewGame newGameDlg;
	
	if (m_game && m_game->thinking)
		return;

	newGameDlg.m_height = m_defaultHeight;
	newGameDlg.m_width = m_defaultWidth;
	newGameDlg.m_depth = m_defaultDepth;
	newGameDlg.m_timeLimit = abs(m_defaultTimeLimit);
	newGameDlg.m_gameLimit = abs(m_defaultGameLimit);
	newGameDlg.m_perMove = (m_defaultGameLimit < 0);

	if (newGameDlg.DoModal() != IDOK)
		return;

	m_dir = dir_invalid;

	if (m_game)
		delete m_game;

	m_game = new DBGame(newGameDlg.m_width, newGameDlg.m_height);
	m_defaultDepth = newGameDlg.m_depth;
	m_defaultTimeLimit = newGameDlg.m_timeLimit;
	m_defaultGameLimit = newGameDlg.m_gameLimit;
	if (newGameDlg.m_perMove)
		m_defaultGameLimit = -m_defaultGameLimit;
	else
		m_defaultTimeLimit = -m_defaultTimeLimit;
	m_defaultHeight = newGameDlg.m_height;
	m_defaultWidth = newGameDlg.m_width;
	m_game->searchDepth = m_defaultDepth;
	m_game->timeLimit = m_defaultTimeLimit;
	m_game->gameLimit = m_defaultGameLimit;
	m_game->hWnd = m_hWnd;

	m_computer = -1;

	Invalidate();
}

void CChildView::GetLine (CPoint p, unsigned &s, unsigned &t, Direction &dir)
{
	int side, x0, y0, u, v;
	float x, y;

	GetCoords(x0, y0, side);

	CRect rect;
	GetClientRect(&rect);
	
	x = (p.x - x0) / (float) side - 1;
	y = (p.y - y0) / (float) side - 1;
	u = floor(x + y);
	v = floor(y - x);

	dir = dir_invalid;
	if ((u ^ v) & 1)
	{
		// horizontal
		s = (u - v + 1) / 2;
		t = (u + v + 1) / 2;
		if (s > 0 && s < (unsigned) m_game->width && t < (unsigned) m_game->height)
			dir = dir_horizontal;
	}
	else
	{
		// vertical
		s = (u - v) / 2;
		t = (u + v + 2) / 2;
		if (s < (unsigned) m_game->width && t > 0 && t < (unsigned) m_game->height)
			dir = dir_vertical;
	}
}

void CChildView::DrawLine (unsigned s, unsigned t, Direction dir, int select)
{
	int x0, y0, side;
	Edge *edge;

	GetCoords(x0, y0, side);

	if (dir == dir_horizontal)
		edge = &m_game->horiz[s][t];
	else
		edge = &m_game->vert[s][t];

	CClientDC dc(this);
	CPen *oldpen = dc.GetCurrentPen();
	CPen *pen;

	if (edge->removed)
	{
		if (select)
			pen = new CPen(PS_SOLID, 0, RGB(64, 255, 64));
		else if (edge->removed == m_game->turn)
			pen = new CPen(PS_SOLID, 0, RGB(255, 0, 255));
		else
			pen = new CPen(PS_SOLID, 0, RGB(0, 0, 0));
	}
	else if (edge->parent && m_internals)
	{
		if (select)
			pen = new CPen(PS_DOT, 0, RGB(255, 64, 64));
		else
			pen = new CPen(PS_DOT, 0, RGB(192, 192, 255));
	}
	else
	{
		if (select)
			pen = new CPen(PS_SOLID, 0, RGB(255, 64, 64));
		else
			pen = new CPen(PS_SOLID, 0, GetSysColor(COLOR_WINDOW));
	}
	dc.SelectObject(pen);

	if (dir == dir_horizontal)
	{
		dc.MoveTo(x0 + s * side + 2, y0 + (t+1) * side);
		dc.LineTo(x0 + (s+1) * side - 2, y0 + (t+1) * side);
	}
	else
	{
		dc.MoveTo(x0 + (s+1) * side, y0 + t * side + 2);
		dc.LineTo(x0 + (s+1) * side, y0 + (t+1) * side - 2);
	}

	dc.SelectObject(oldpen);
	delete pen;
}

void CChildView::OnMouseMove(UINT nFlags, CPoint point) 
{
	unsigned s, t;
	Direction dir;

	if (m_game)
	{
		GetLine(point, s, t, dir);
		if (dir != m_dir || s != m_s || t != m_t)
		{
			if (m_dir != dir_invalid)
				DrawLine(m_s, m_t, m_dir);
			GetLine(point, m_s, m_t, m_dir);
			if (m_dir != dir_invalid)
				DrawLine(m_s, m_t, m_dir, 1);
		}
	}

	CWnd ::OnMouseMove(nFlags, point);
}

UINT DoMove2 (void *g)
{
	DBGame *game = (DBGame *) g;

	game->MyMove();
	game->stop = 1;
	game->thinking = 0;
	::PostMessage(game->hWnd, WM_CHAR, 18, 0);

	return 0;
}

UINT DoMove (void *g)
{
	DBGame *game = (DBGame *) g;

	game->thinking = 1;
	game->stop = 0;

	AfxBeginThread(DoMove2, g);

	int timelimit;
	if (game->gameLimit < 0)
		timelimit = game->timeLimit;
	else
	{
#ifdef HASH
		int totalmoves = game->width * game->height;
		int a = totalmoves * 14 / 36;
		int b = totalmoves * 22 / 36;
		if (!a)
			a = 1;
		if (b < a + 2)
			b = a + 2;
		int start = game->gameLimit * 9;
		int end = game->gameLimit * 9;
		int lastmid = game->gameLimit * 6;
		int middle = 60 * game->gameLimit - start - end - lastmid;

		int movenum = game->nummoves + 1;
		if (movenum < a)
			timelimit = start / (a / 2);
		else if (movenum < b)
		{
			int midsize = (b - a) / 2 - 1;
			int imid = (movenum - a) / 2;
			if (imid == midsize)
				timelimit = lastmid;
			else
			{
				midsize = midsize * (midsize + 1) / 2;
				timelimit = middle * (imid + 1) / midsize;
			}
		}
		else
			timelimit = 2 * end / (totalmoves - b);
#else
		int totalmoves = game->width * game->height / 2;
		int movenum = (game->nummoves / 2) + 1;
		int firsthalf = game->gameLimit > 2 ? 60 : 20;
		int secondhalf = 60 * game->gameLimit - firsthalf;
		if (movenum < totalmoves / 2)
			timelimit = firsthalf * 2 / totalmoves;
		else if (totalmoves > 10)
			timelimit = 2 * secondhalf / (totalmoves - 10);
		else
			timelimit = secondhalf / 2;
		if (!timelimit)
			timelimit = 1;
#endif
	}

	if (!timelimit)
		return 0;

	for (int i = 0 ; i < timelimit * 4 ; i++)
	{
		Sleep(250);
		if (game->stop)
			break;
	}

	game->stop = 1;
	game->thinking = 0;
	::PostMessage(game->hWnd, WM_CHAR, 18, 0);

	return 0;
}

void CChildView::OnLButtonDown(UINT nFlags, CPoint point) 
{	
	unsigned s, t;
	Direction dir;
	int f;

	if (m_game)
	{
		if (m_game->thinking)
			return;

		GetLine(point, s, t, dir);
		if (dir != dir_invalid)
		{
			f = m_game->AddLine(s, t, dir);
			Invalidate();

			if (!f && m_autoPlay)
			{
				m_computer = m_game->nummoves & 1;
				AfxBeginThread(DoMove, m_game);
			}
		}
	}
	CWnd::OnLButtonDown(nFlags, point);
}

void CChildView::OnMymove() 
{
	// TODO: Add your command handler code here

	if (m_game && !m_game->thinking)
	{
		if (m_autoPlay)
			m_computer = m_game->nummoves & 1;
		AfxBeginThread(DoMove, m_game);
	}
}

void CChildView::OnSettings() 
{
	CDlgSettings settingsDlg;

	if (m_game && m_game->thinking)
		return;

	settingsDlg.m_depth = m_defaultDepth;	
	settingsDlg.m_timeLimit = abs(m_defaultTimeLimit);
	settingsDlg.m_gameLimit = abs(m_defaultGameLimit);
	settingsDlg.m_perMove = (m_defaultGameLimit < 0);
	settingsDlg.m_name1 = m_name1;
	settingsDlg.m_name2 = m_name2;

	if (settingsDlg.DoModal() != IDOK)
		return;

	m_defaultDepth = settingsDlg.m_depth;
	m_defaultTimeLimit = settingsDlg.m_timeLimit;
	m_defaultGameLimit = settingsDlg.m_gameLimit;
	m_name1 = settingsDlg.m_name1;
	m_name2 = settingsDlg.m_name2;
	if (settingsDlg.m_perMove)
		m_defaultGameLimit = -m_defaultGameLimit;
	else
		m_defaultTimeLimit = -m_defaultTimeLimit;
	if (m_game)
	{
		m_game->searchDepth = settingsDlg.m_depth;
		m_game->timeLimit = m_defaultTimeLimit;
		m_game->gameLimit = m_defaultGameLimit;
		Invalidate();
	}
}

void CChildView::OnEval() 
{
	if (m_game)
	{
		m_game->leaves = 0;
		int val = m_game->Evaluate(INF, -INF, 0);

		TRACE("val = %3d   leaves = %d\n", val, m_game->leaves);
	}
}

void CChildView::OnUndo() 
{
	if (m_game)
	{
		if (m_game->thinking)
			return;

		m_game->Undo();
		Invalidate();
	}
}

void CChildView::OnAutoPlay() 
{
	m_autoPlay = !m_autoPlay;
	if (m_computer != -1)
	{
		m_computer = -1;
		Invalidate();
	}
}

void CChildView::OnUpdateAutoPlay(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(m_autoPlay);	
}

void CChildView::OnInternals() 
{
	m_internals = !m_internals;
	Invalidate();
}

void CChildView::OnUpdateInternals(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(m_internals);	
}

void CChildView::OnUpdateEval(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(0);	
}

void CChildView::OnChar(UINT nChar, UINT nRepCnt, UINT nFlags)
{
	if (nChar == 18)
		Invalidate();

	CWnd::OnChar(nChar, nRepCnt, nFlags);
}

void CChildView::OnForward() 
{
	if (m_game)
	{
		if (m_game->thinking)
			return;

		if (m_game->numEdgesRemoved < m_game->maxEdgesRemoved)
		{
			m_game->Redo();
			Invalidate();
		}
	}
}

void CChildView::OnUpdateForward(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(m_game && m_game->numEdgesRemoved < m_game->maxEdgesRemoved && !m_game->thinking);	
}

void CChildView::OnUpdateUndo(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(m_game && m_game->numEdgesRemoved && !m_game->thinking);	
}

void CChildView::OnUpdateMymove(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(m_game && !m_game->thinking);
}

void CChildView::OnFileOpen() 
{
	CFileDialog dlg(TRUE, "dbl", NULL, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT, "Dabble files (*.dbl)|*.dbl||", this);
	char dir[128];

	GetCurrentDirectory(128, dir);
	dlg.m_ofn.lpstrInitialDir = dir;
	
	if (m_game && m_game->thinking)
		return;

	if (dlg.DoModal() == IDOK)
	{
		CString name = dlg.GetPathName();
		FILE *f = fopen(name, "r");
		if (!f)
			MessageBox("Could not open file", "Error");
		else
		{
			int w, h;
			fscanf(f, "%d, %d\n", &w, &h);

			if (w > MAX_SIZE || h > MAX_SIZE)
			{
#ifdef HASH
				MessageBox("Saved game too large (try using dabble_nohash.exe)");
#else
				MessageBox("Saved game too large");
#endif
				fclose(f);
				return;
			}

			if (m_game)
				delete m_game;

			m_game = new DBGame(w, h);
			m_game->searchDepth = m_defaultDepth;
			m_game->timeLimit = m_defaultTimeLimit;
			m_game->gameLimit = m_defaultGameLimit;
			m_game->hWnd = m_hWnd;

			int x1, x2, y1, y2;
			Coords c;
			while (fscanf(f, "(%d, %d) - (%d, %d)", &x1, &y1, &x2, &y2) && !feof(f))
			{
				c.x1 = x1;
				c.x2 = x2;
				c.y1 = y1;
				c.y2 = y2;
				m_game->rgEdgeRemoved[m_game->maxEdgesRemoved] = c;
				m_game->maxEdgesRemoved++;
				fscanf(f, "\n");
			}
			fclose(f);
			Invalidate();
		}
	}
}

void CChildView::OnFileSave() 
{
	CFileDialog dlg(FALSE, "dbl", NULL, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT, "Dabble files (*.dbl)|*.dbl||", this);
	char dir[128];

	if (!m_game || m_game->thinking)
		return;
	
	GetCurrentDirectory(128, dir);
	dlg.m_ofn.lpstrInitialDir = dir;
	
	if (dlg.DoModal() == IDOK)
	{
		CString name = dlg.GetFileName();
		FILE *f = fopen(name, "w");
		if (!f)
			MessageBox("Could not open file", "Error");
		else
		{
			fprintf(f, "%d, %d\n", m_game->width, m_game->height);
			for (int i = 0 ; i < m_game->maxEdgesRemoved ; i++)
			{
				Coords c = m_game->rgEdgeRemoved[i];
				fprintf(f, "(%d, %d) - (%d, %d)\n", c.x1, c.y1, c.x2, c.y2);
			}
			fclose(f);
		}
	}	
}

void CChildView::OnUpdateStop(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(m_game && m_game->thinking);	
}

void CChildView::OnStop() 
{
	if (m_game && m_game->thinking)
		m_game->stop = 1;
}

void CChildView::OnUpdateNewGame(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(!m_game || !m_game->thinking);	
}

void CChildView::OnUnpause() 
{
	if (m_game)
		m_game->m_pause = 0;
}

void CChildView::OnShowBeta() 
{
	m_showBeta = !m_showBeta;	
	Invalidate();
}

void CChildView::OnUpdateShowBeta(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(m_showBeta);	
}
